/*
# Fish & Shark
# Copyright (C) 2012 Sebastian Rothe & Alexander Lambertz
#
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with this program;
# if not, see <http://www.gnu.org/licenses/>.
#
# Abstract:
# This simulation is a modification of a Wator Theory. It simulates fishs, sharks and plankton.
# Fishs die and respawn after a certain period of time. Sharks eat fishs and duplicate themselves.
# Plankton is just growing all the time. This simulation was parallized by using the MPI Framework.
#
# Please notice that you might need to edit 3 times the directory for the output in the debug.h file.
#
*/

// -----------------------------
// ## Includes ##
// -----------------------------
#include <stdio.h>
#include <stdlib.h>
// Include Debugging-File
#include "debug.h"

// For Real Randomized Numbers
#include <time.h>

// MPI-Library
#include <mpi.h>

// -----------------------------
// ## Declare Local-Constants ##
// -----------------------------
// Field-Size
#define iFieldSizeX 1000
#define iFieldSizeY 1000

#define iIntervalCountMax 80

// Values on the field
#define iMarkEmpty 0
#define iMarkFish 1
#define iMarkShark 2
#define iMarkPlankton 3

// Debugging
// Initializse Debug-Process
const int iDebugging = 0;
const int iDebugProcess = 1;

// Respawntimes
const int iRespawnMinFish = 5;
const int iRespawnMaxFish = 10;

const int iRespawnMinPlankton = 5;
const int iRespawnMaxPlankton = 15;

const int iLifetimeMin = 40;
const int iLifetimeMax = 150;

const int iInitialFishAmount = 750000;
const int iInitialSharkAmount = 1000;
const int iInitialPlanktonAmount = 10000;

const int iRndToZero = 5;

// Counting Elements for stopping loops if no more Elements are expected to appear.
int iAmountFish=0;
int iAmountShark=0;
int iAmountPlankton=0;

// -----------------------------
// ## Functions ##
// -----------------------------


// Generate a random Position on the field
// in the range of the X and Y-Values for the field
// Needs Pointer to y and x to return Values.
void rndPosition(int *y,int *x)
{
    *y = rand() % iFieldSizeY;
    *x = rand() % iFieldSizeX;
    
    // Debugging
    // printf("Generated Y-Value: %d\n", *y);
    // printf("Generated X-Value: %d\n", *x);
}

// Generate a random Move
// two Value from -1 to 1
void rndMove(int *y,int *x)
{

    // Prevent Diagonal
    int iRand = (rand() % 5);
    
    switch (iRand)
    {
        case 0:
            *x = 0;
            *y = 0;
            break;
        case 1:
            *x = -1;
            *y = 0;
            break;
        case 2:
            *x = 1;
            *y = 0;
            break;
        case 3:
            *x = 0;
            *y = 1;
            break;
        case 4:
            *x = 0;
            *y = -1;
            break;
    }
}

// Generate random Respawn Interval
int rndRespawn(int actual, int min, int max)
{
    int iReturnValue = (rand() % (max-min))+actual+min;
    
    return iReturnValue;
    
    // Debugging
    // printf("Generated Respawn-Value: %d\n", iReturnValue);
}

// Check if Position is free
// 1 = Free
// 0 = Blocked
//
// Please notice if position doesn't change the position will be returned as blocked.
int checkIfFree(int ***array, int *aBorderTop, int *aBorderRight, int *aBorderBottom, int *aBorderLeft, int y, int x)
{
    if (y == -1)
        if (aBorderTop[(x)] == 0)
            return 1;
        else
            return 0;
    else if (x == -1)
        if (aBorderLeft[(y)] == 0)
            return 1;
        else
            return 0;
    else if (y == iFieldSizeY)
        if (aBorderBottom[(x)] == 0)
            return 1;
        else
            return 0;
    else if (x == iFieldSizeX)
        if (aBorderRight[(y)] == 0)
            return 1;
        else
            return 0;
    else if ( array[y][x][0] == 0 )
        return 1;
    else
        return 0;
}

// Checks if there is a fish around the return absolute position
// 1 = Yes there is a fish (parameters are rewriten with the fish's Coordinates
// 0 = No fish available
//
// Returns x and Y with Position of Fish

int checkIfFishAround(int ***array, int *aBorderTop, int *aBorderRight, int *aBorderBottom, int *aBorderLeft, int *y, int *x)
{
    
    //x=0 y=1
    if (*y+1>=iFieldSizeY)
    {
        if (aBorderBottom[(*x)] == iMarkFish)
        {
            *y=*y+1;
            *x=*x;
            return 1;
        }
    } else {
        if (array[*y+1][*x][0] == iMarkFish)
        {
            *y=*y+1;
            *x=*x;
            return 1;
        }
    }
    // x=0 y=-1
    if (*y-1<0)
    {
        if (aBorderTop[(*x)] == iMarkFish)
        {
            *y=*y-1;
            *x=*x;
            return 1;
        }
    } else {
        if (array[*y-1][*x][0] == iMarkFish)
        {
            *y=*y-1;
            *x=*x;
            return 1;
        }
    }
    
    // x=-1 y=0
    if (*x+1>=iFieldSizeX)
    {
        if (aBorderRight[(*y)] == iMarkFish)
        {
            *y=*y;
            *x=*x+1;
            return 1;
        }
    } else {
        if (array[*y][*x+1][0] == iMarkFish)
        {
            *y=*y;
            *x=*x+1;
            return 1;
        }
    }
    // x=1 y=0
    if (*x-1<0)
    {
        if (aBorderLeft[(*y)] == iMarkFish)
        {
            *y=*y;
            *x=*x-1;
            return 1;
        }
    } else {
        if (array[*y][*x-1][0] == iMarkFish)
        {
            *y=*y;
            *x=*x-1;
            return 1;
        }
    }
    return 0;
}

// Reset touched this round
//
// Function iterates through whole array and resets value 2 of Array
void resetTouched(int ***array)
{
    for (int i=0; i<iFieldSizeY; i++) {
        for (int j=0; j<iFieldSizeX; j++) {
            array[i][j][2] = 0;
        }
    }
}

// Reset all Values for Sector given as Parameter
//
void resetSector(int ***array, int y, int x)
{
    array[y][x][0] = 0;
    array[y][x][1] = 0;
    array[y][x][2] = 0;
}


// Move all Values from one Sector to another
void moveValues(int ***array, int oldY, int oldX, int newY, int newX)
{
    array[newY][newX][0] = array[oldY][oldX][0];
    array[newY][newX][1] = array[oldY][oldX][1];
    array[newY][newX][2] = 1;
    
    // Only Perform Move if real Movement
    if ( (oldY != newY) || (oldX != newX))
    {
        resetSector(array, oldY, oldX);
    }
}

// Generate new Lifetime
int generateLifeTime(int actual)
{
    return  ((rand() % (iLifetimeMax-iLifetimeMin))+iLifetimeMin+actual);
}

// Set a Fish on the delivered Position
// Please be careful that the position is free!
void setFish(int ***array, int y, int x, int actual)
{
    array[y][x][0] = iMarkFish;
    array[y][x][1] = generateLifeTime(actual);
    array[y][x][2] = 1; // Set Sector TOUCHED for this round
    iAmountFish++;
}
// Set a Shark on the delivered Position
// Please be careful that the position is free!
void setShark(int ***array, int y, int x, int actual)
{
    array[y][x][0] = iMarkShark;
    array[y][x][1] = generateLifeTime(actual);
    array[y][x][2] = 1; // Set Sector TOUCHED for this round
    iAmountShark++;
}
// Set a Plankton
void setPlankton(int ***array, int y, int x, int actual)
{
    array[y][x][0] = iMarkPlankton;
    array[y][x][1] = rndRespawn(actual, iRespawnMinPlankton, iRespawnMaxPlankton);
    array[y][x][2] = 1; // Set Sector TOUCHED for this round
    iAmountPlankton++;
}

// Set Value
void setValue(int ***array, int y, int x, int value, int actual)
{
    if (value == iMarkFish)
    {
        setFish(array, y, x, actual);
    }
    else if (value == iMarkShark)
    {
        setShark(array, y, x, actual);
    }
    else if (value == iMarkPlankton)
    {
        setPlankton(array, y, x, actual);
    }

    //Depricated: array[y][x][0] = value;
}


// Check if Position is in Border-Area
// 0 := No
// 1 := Top
// 2 := Right
// 3 := Botton
// 4 := Left
int checkIfBorder(int y, int x)
{
     if ( (y == 0) || (y == -1) )
        return 1;
    else if ( (y == (iFieldSizeY-1)) || (y == (iFieldSizeY)) )
        return 3;
    else if ( (x == 0) || (x == -1) )
        return 4;
    else if ( (x == (iFieldSizeX-1)) || (x == (iFieldSizeX)) )
        return 2;
    else
        return 0;
}

// Check if Position is in other Process
// 1 = Yes
// 0 = No
int checkIfOtherProcess(int y, int x)
{
    if ( (y == -1) || (y == (iFieldSizeY)) )
        return 1;
    else if ( (x == -1) || (x == (iFieldSizeX)) )
        return 1;
    else
        return 0;
}

               
// Function to deal with Border-Movement and Objects moving to other Process
//
//
void specialMovement(int ***array, int *aBorderTop, int *aBorderRight, int *aBorderBottom, int *aBorderLeft, int *data, int from, int actual)
{
    
    // Specification for from:
    // 1 = Top
    // 2 = Right
    // 3 = Bottom
    // 4 = Left
    
    // Specification for Data Array:
    // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    // !! Be Carful: Position is from Process sending the Data !!
    // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    // 0 = OLD-Y
    // 1 = OLD-X
    // 2 = NEW-Y
    // 3 = NEW-X
    // 4 = Type
    
    if (iDebugging == 2)
        printf("Special Movement: %d/%d/%d/%d/%d/%d\n", from, data[0], data[1], data[2], data[3], data[4]);
    
    // *TOP*
    if (from == 3)
    {
        // Border Area
        if (data[2] == (iFieldSizeY-1))
        {
            // Check if Items has been in Border before!!
            if (data[0] == (iFieldSizeY-1))
            {
                aBorderTop[(data[1])] = 0;
            }
            aBorderTop[(data[3])] = data[4];
        // Process Change
        } else if (data[2] == iFieldSizeY)
        {
            if (data[0] == (iFieldSizeY-1))
            {
                aBorderTop[(data[1])] = 0;
            }
            setValue(array, 0, data[3], data[4], actual);
        } else if (data[2] == -2)
        {
            if (data[0] == (iFieldSizeY-1))
            {
                aBorderTop[(data[1])] = 0;
            }
        }
    } else if (from == 4)
    {
        // Border Area
        if (data[3] == (iFieldSizeX-1))
        {
            // Check if Items has been in Border before!!
            if (data[1] == (iFieldSizeX-1))
            {
                aBorderRight[(data[0])] = 0;
            }
            aBorderRight[(data[2])] = data[4];
            // Process Change
        } else if (data[3] == iFieldSizeX)
        {
            if (data[1] == (iFieldSizeX-1))
            {
                aBorderRight[(data[0])] = 0;
            }
            setValue(array, data[2], 0, data[4], actual);
        } else if (data[3] == -2)
        {
            if (data[1] == (iFieldSizeX-1))
            {
                aBorderRight[(data[0])] = 0;
            }
        }
    } else if (from == 1)
    {
        // Border Area
        if (data[2] == 0)
        {
            // Check if Items has been in Border before!!
            if (data[0] == 0)
            {
                aBorderBottom[(data[1])] = 0;
            }
            aBorderBottom[(data[3])] = data[4];
            // Process Change
        } else if (data[2] == -1)
        {
            if (data[0] == 0)
            {
                aBorderBottom[(data[1])] = 0;
            }
            setValue(array, 0, data[3], data[4], actual);
        } else if (data[2] == -2)
        {
            if (data[0] == 0)
            {
                aBorderBottom[(data[1])] = 0;
            }
        }
    } else if (from == 2)
    {
        // Border Area
        if (data[3] == 0)
        {
            // Check if Items has been in Border before!!
            if (data[1] == 0)
            {
                aBorderLeft[(data[0])] = 0;
            }
            aBorderLeft[(data[2])] = data[4];
            // Process Change
        } else if (data[3] == -1)
        {
            if (data[1] == 0)
            {
                aBorderLeft[(data[0])] = 0;
            }
            setValue(array, data[2], 0, data[4], actual);
        } else if (data[3] == -2)
        {
            if (data[1] == 0)
            {
                aBorderLeft[(data[0])] = 0;
            }
        }
    }
}

void freeMemory(int ***array, int *aBorderTop, int *aBorderRight, int *aBorderBottom, int *aBorderLeft)
{
    if (array!=NULL)
        free(array);
    if (aBorderTop!=NULL)
        free(aBorderTop);
    if (aBorderRight!=NULL)
        free(aBorderRight);
    if (aBorderBottom!=NULL)
        free(aBorderBottom);
    if (aBorderLeft!=NULL)
        free(aBorderLeft);
}

// -----------------------------
// ## START ##
// -----------------------------
int main(argc, argv)
int        argc;
char       *argv[];
{
    clock_t cPrgstart, cPrgend;
    cPrgstart=clock();
    
    // Buffer for Request-Status
    MPI_Request mpiRRequest;
   
    // Initialize MPI
    MPI_Init(&argc,&argv);
    
    // Define variables for MPI-Neighbors
    int mpiITop;
    int mpiIRight;
    int mpiIBottom;
    int mpiILeft;
    
    // Array of 5 Values to transmit a movement between the different processes
    // Specification:
    // 0 = OLD-Y
    // 1 = OLD-X
    // 2 = NEW-Y
    // 3 = NEW-X
    // 4 = Type
    int *mpiAData;
    mpiAData=(int*) malloc(sizeof(*mpiAData)*5);
    
    // Variables for own Status
    int mpiIMe;
    int mpiISize;
    
    int mpiIProcessPerline;
    
    // MPI: Determine - Who am I?
    MPI_Comm_rank (MPI_COMM_WORLD, &mpiIMe);
    MPI_Comm_size (MPI_COMM_WORLD, &mpiISize);
    
    if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
        printf("MPI Initialisation finished\n");
    
    // Determine how many processes per line will be initialized
    if (mpiISize%4==0)
        mpiIProcessPerline=4;
    else if (mpiISize%3==0)
        mpiIProcessPerline=3;
    else if (mpiISize%2==0)
        mpiIProcessPerline=2;
    else if (mpiISize%1==0)
        mpiIProcessPerline=1;
    
    // Debugging Processes per line
    if (mpiIMe==iDebugProcess && iDebugging>0)
        printf("I will place %i Processes per line \n\n", mpiIProcessPerline);
    
    
    // ---------------------------------------------
    // ## Determine Neighborhood for each Process ##
    // ---------------------------------------------
    
    // Determine Neighbors
    //
    // Visualisaton 12 Processes
    // |-|-|-|-|
    // |0|1|2|3|
    // |-|-|-|-|
    // |4|5|6|7|
    // |-|-|-|-|
    // |8|9|10|11|
    // |-|-|-|-|
    //
    
    // TOP
    if (0==(mpiIMe/mpiIProcessPerline)) // Bottom & Rest
        mpiITop=((mpiISize/mpiIProcessPerline)-1)*mpiIProcessPerline+(mpiIMe%mpiIProcessPerline);
    else if ((mpiISize/mpiIProcessPerline)>(mpiIMe/mpiIProcessPerline)) // Top
        mpiITop=mpiIMe-mpiIProcessPerline;
    
    // Right
    if ((mpiIMe%mpiIProcessPerline)<(mpiIProcessPerline-1))
        mpiIRight=(mpiIMe+1);
    else 
        mpiIRight=(mpiIMe-(mpiIProcessPerline-1));
    
    // Bottom
    if (((mpiISize/mpiIProcessPerline)-1)==(mpiIMe/mpiIProcessPerline)) // Bottom
        mpiIBottom=(mpiIMe%mpiIProcessPerline);
    else if ((mpiISize/mpiIProcessPerline)>(mpiIMe/mpiIProcessPerline)) // Top & Rest
        mpiIBottom=mpiIMe+mpiIProcessPerline;
    
    // Left
    if ((mpiIMe%mpiIProcessPerline)>0)
        mpiILeft=(mpiIMe-1);
    else 
        mpiILeft=(mpiIMe+(mpiIProcessPerline-1));
    

    if (mpiIMe==iDebugProcess && iDebugging>0)
        printf("Top: %i - Right: %i - Bottom: %i - Left: %i\n\n", mpiITop, mpiIRight, mpiIBottom, mpiILeft);
    
    // Real Randoms by using MPI-Process-No.
    srand ( time(NULL) * mpiIMe );
    
    if (iDebugging == 1)
        printf("%d: %d\n",mpiIMe,rand());
    

    // Variable to summarize all Amounts (Fish, Shark, Plankton)
    int iAmountAll;
    int iAmountProcessed;
    
    // RespawnInterval for Fish
    int iRespawnFish = 0;
    
    int iOldValue = 0;
    
    int iRnds = 0;
    
    // MPI Stuff
    int mpiIBorderTouched;
    MPI_Status  mpiStat;
    
    // ---------------------------------------------
    // ##       Initialize Simulation Field       ##
    // ---------------------------------------------
    int ***aField;

    aField = malloc(sizeof(int *) * iFieldSizeY);
    
    if (aField == NULL)
    {
        printf("Memory Issue");
        exit(1);
    }
        
    for (int y=0;y<iFieldSizeY;y++)
    {
        aField[y] = malloc (sizeof(int *) * iFieldSizeX);
        // Set all values to 0
        for (int x=0;x<iFieldSizeX;x++)
        {
            aField[y][x] = malloc (sizeof(int) * 3);
            // Initialize 3 int-Values for each Field-Position
            // Position 0 := Value of Sector
            // Position 1 := DieTime
            // Position 2 := Touched this round
            aField[y][x][0] = 0;
            aField[y][x][1] = -1;
            aField[y][x][2] = 0;
        }
    }
    
    if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
        printf("Inistialized Field\n");
    
    // ---------------------------------------------
    // ##          Initialize Border Area         ##
    // ---------------------------------------------
    // Arrays for Border-Area
    int *aBorderTop;
    int *aBorderRight;
    int *aBorderLeft;
    int *aBorderBottom;
    
    // +2 because Boder area is 2 Steps bigger then the normal area!!
    // TODO: Redundance; not needed but not critical
    aBorderTop = malloc(sizeof(int) * (iFieldSizeX));
    aBorderBottom = malloc(sizeof(int) * (iFieldSizeX));
    
    if (aBorderTop == NULL || aBorderBottom == NULL)
    {
        printf("Memory Issue");
        // Free Memory
        freeMemory(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft);
        exit(1);
    }
    
    for (int x=0;x<(iFieldSizeX);x++)
    {
        aBorderTop[x] = 0;
        aBorderBottom[x] = 0;
    }
    
    aBorderLeft = malloc(sizeof(int) * (iFieldSizeY));
    aBorderRight = malloc(sizeof(int) * (iFieldSizeY));
    
    if (aBorderLeft == NULL || aBorderRight == NULL)
    {
        printf("Memory Issue");
        // Free Memory
        freeMemory(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft);
        exit(1);
    }
    
    for (int y=0;y<(iFieldSizeY);y++)
    {
        aBorderLeft[y] = 0;
        aBorderRight[y] = 0;
    }
    
    if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
        printf("Initialized Border\n");
    
    // Variables for Position
    int x,y;
    
    // ---------------------------------------------
    // ##            Set initial Fishs            ##
    // ---------------------------------------------
    for (int i=0; i<iInitialFishAmount; i++) {
        
        do {
            rndPosition(&y, &x);
        } while (checkIfFree(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, y, x)==0);
    
        setFish(aField, y, x, 0);
        
        if (checkIfBorder(y, x)>0)
        {
            if (iDebugging == 2 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                printf("Border A1\n");
            
            int temp = checkIfBorder(y, x);
            mpiAData[0] = -2;
            mpiAData[1] = -2;
            mpiAData[2] = y;
            mpiAData[3] = x;
            mpiAData[4] = iMarkFish;
            
            // Send Data to other Processes
            if (temp == 1)
            {
                MPI_Isend(mpiAData, 5, MPI_INT , mpiITop, 0, MPI_COMM_WORLD, &mpiRRequest);
                MPI_Wait(&mpiRRequest,&mpiStat);
            }
            
            if (temp == 2)
            {
                MPI_Isend(mpiAData, 5, MPI_INT , mpiIRight, 0, MPI_COMM_WORLD, &mpiRRequest);
                MPI_Wait(&mpiRRequest,&mpiStat);
            }
            
            if (temp == 3)
            {
                MPI_Isend(mpiAData, 5, MPI_INT , mpiIBottom, 0, MPI_COMM_WORLD, &mpiRRequest);
                MPI_Wait(&mpiRRequest,&mpiStat);
            }
            
            if (temp == 4)
            {
                MPI_Isend(mpiAData, 5, MPI_INT , mpiILeft, 0, MPI_COMM_WORLD, &mpiRRequest);
                MPI_Wait(&mpiRRequest,&mpiStat);
            }
        }
        
    }
    
    iRespawnFish = rndRespawn(0, iRespawnMinFish, iRespawnMaxFish);
    
    if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
        printf("Setted Initial Fish\n");
    
    // ---------------------------------------------
    // ##           Set initial Sharks            ##
    // ---------------------------------------------
    for (int i=0; i<iInitialSharkAmount; i++) {
        
        do {
            rndPosition(&y, &x);
        } while (checkIfFree(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, y, x)==0);
    
        setShark(aField, y, x, 0);
        
        if (checkIfBorder(y, x)>0)
        {
            if (iDebugging == 2 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                printf("Border A2\n");
            
            int temp = checkIfBorder(y, x);
            mpiAData[0] = -2;
            mpiAData[1] = -2;
            mpiAData[2] = y;
            mpiAData[3] = x;
            mpiAData[4] = iMarkShark;
            
            // Send Data to other Processes
            if (temp == 1)
            {
                MPI_Isend(mpiAData, 5, MPI_INT , mpiITop, 0, MPI_COMM_WORLD, &mpiRRequest);
                MPI_Wait(&mpiRRequest,&mpiStat);
            }
            
            if (temp == 2)
            {
                MPI_Isend(mpiAData, 5, MPI_INT , mpiIRight, 0, MPI_COMM_WORLD, &mpiRRequest);
                MPI_Wait(&mpiRRequest,&mpiStat);
            }
            
            if (temp == 3)
            {
                MPI_Isend(mpiAData, 5, MPI_INT , mpiIBottom, 0, MPI_COMM_WORLD, &mpiRRequest);
                MPI_Wait(&mpiRRequest,&mpiStat);
            }
            
            if (temp == 4)
            {
                MPI_Isend(mpiAData, 5, MPI_INT , mpiILeft, 0, MPI_COMM_WORLD, &mpiRRequest);
                MPI_Wait(&mpiRRequest,&mpiStat);
            }
        }
        
    }
    
    if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
        printf("Setted Initial Shark\n");
    
    // ---------------------------------------------
    // ##           Set initial Plankton           ##
    // ---------------------------------------------
    for (int i=0; i<iInitialPlanktonAmount; i++) {
        
        do {
            rndPosition(&y, &x);
        } while (checkIfFree(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, y, x)==0);
    
        setPlankton(aField, y, x, 0);
        
        if (checkIfBorder(y, x)>0)
        {
            if (iDebugging == 2 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                printf("Border A3\n");
            
            int temp = checkIfBorder(y, x);
            mpiAData[0] = -2;
            mpiAData[1] = -2;
            mpiAData[2] = y;
            mpiAData[3] = x;
            mpiAData[4] = iMarkPlankton;
            
            // Send Data to other Processes
            if (temp == 1)
            {
                MPI_Isend(mpiAData, 5, MPI_INT , mpiITop, 0, MPI_COMM_WORLD, &mpiRRequest);
                MPI_Wait(&mpiRRequest,&mpiStat);
            }
            
            if (temp == 2)
            {
                MPI_Isend(mpiAData, 5, MPI_INT , mpiIRight, 0, MPI_COMM_WORLD, &mpiRRequest);
                MPI_Wait(&mpiRRequest,&mpiStat);
            }
            
            if (temp == 3)
            {
                MPI_Isend(mpiAData, 5, MPI_INT , mpiIBottom, 0, MPI_COMM_WORLD, &mpiRRequest);
                MPI_Wait(&mpiRRequest,&mpiStat);
            }
            
            if (temp == 4)
            {
                MPI_Isend(mpiAData, 5, MPI_INT , mpiILeft, 0, MPI_COMM_WORLD, &mpiRRequest);
                MPI_Wait(&mpiRRequest,&mpiStat);
            }
        }
        
    }
    
    if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
        printf("Setted Initial Plankton\n");
    
    iAmountAll = -1;
    
    for (int iIntervalCount=0; iIntervalCount<iIntervalCountMax; iIntervalCount++) {
        
        if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
            printf("Started Round %d\n", mpiIMe);
        
        // ---------------------------------------------
        // ##         Respawn Fish if needed          ##
        // ---------------------------------------------
        if (iIntervalCount == iRespawnFish)
        {
            do {
                rndPosition(&y, &x);
            } while ( checkIfFree(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, y, x)==0);
            
            setFish(aField, y, x, iIntervalCount);
            
            if (checkIfBorder(y, x)>0)
            {
                
                // Delete from other processes if Border
                int temp = checkIfBorder(y, x);
                
                mpiAData[0] = -2;
                mpiAData[1] = -2;
                mpiAData[2] = y;
                mpiAData[3] = x;
                mpiAData[4] = iMarkFish;
                
                // Send Data to other Processes
                if (temp == 1)
                {
                    MPI_Isend(mpiAData, 5, MPI_INT , mpiITop, 0, MPI_COMM_WORLD, &mpiRRequest);
                    MPI_Wait(&mpiRRequest,&mpiStat);
                }
                
                if (temp == 2)
                {
                    MPI_Isend(mpiAData, 5, MPI_INT , mpiIRight, 0, MPI_COMM_WORLD, &mpiRRequest);
                    MPI_Wait(&mpiRRequest,&mpiStat);
                }
                
                if (temp == 3)
                {
                    MPI_Isend(mpiAData, 5, MPI_INT , mpiIBottom, 0, MPI_COMM_WORLD, &mpiRRequest);
                    MPI_Wait(&mpiRRequest,&mpiStat);
                }
                
                if (temp == 4)
                {
                    MPI_Isend(mpiAData, 5, MPI_INT , mpiILeft, 0, MPI_COMM_WORLD, &mpiRRequest);
                    MPI_Wait(&mpiRRequest,&mpiStat);
                }
            }
            
            iRespawnFish = rndRespawn(iIntervalCount, iRespawnMinFish, iRespawnMaxFish);
        }
        
        if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
            printf("Respawned a fish\n");
        
        // New Picture -> New Amount
        iAmountProcessed=0;
        
        // ---------------------------------------------
        // ##      Iterate through the field          ##
        // ---------------------------------------------
        
        for (int i=0; (i<iFieldSizeY) && (iAmountAll>=iAmountProcessed); i++) {
            for (int j=0; (j<iFieldSizeX) && (iAmountAll>=iAmountProcessed); j++) {
        //for (int i=0; (i<iFieldSizeY); i++) {
        //    for (int j=0; (j<iFieldSizeX); j++) {
                
                if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                    printf("Amounts1: %d/%d\n",iAmountProcessed,iAmountAll);
                
                if (iDebugging == 4 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                    printf("Checking %d/%d (Data: %d/%d/%d\n",i,j,aField[i][j][0],aField[i][j][1],aField[i][j][2]);
                
                // ---------------------------------------------
                // ##         MPI Receive from all            ##
                // ##          Neighbor Processes             ##
                // ---------------------------------------------
                
                // MPI: Check if another Process wants to transmit a Border-Movement
                MPI_Iprobe(mpiITop, 0, MPI_COMM_WORLD, &mpiIBorderTouched, &mpiStat);
                
                // If there is Data to recieve => Recieve it
                while (mpiIBorderTouched == 1) {
                    if (iDebugging == 2 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                        printf("Received from Top\n");
                    
                    // Recieve Stuff
                    MPI_Irecv(mpiAData, 5, MPI_INT, mpiITop, 0, MPI_COMM_WORLD, &mpiRRequest);
                    specialMovement(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, mpiAData, 1, iIntervalCount);
                    
                    MPI_Iprobe(mpiITop, 0, MPI_COMM_WORLD, &mpiIBorderTouched, &mpiStat);
                }
                
                // MPI: Check if another Process wants to transmit a Border-Movement
                MPI_Iprobe(mpiIRight, 0, MPI_COMM_WORLD, &mpiIBorderTouched, &mpiStat);
                
                // If there is Data to recieve => Recieve it
                while (mpiIBorderTouched == 1) {
                    
                    if (iDebugging == 2 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                        printf("Received from Right\n");
                    
                    // Recieve Stuff
                    MPI_Irecv(mpiAData, 5, MPI_INT, mpiIRight, 0, MPI_COMM_WORLD, &mpiRRequest);
                    specialMovement(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, mpiAData, 2, iIntervalCount);
                    
                    MPI_Iprobe(mpiIRight, 0, MPI_COMM_WORLD, &mpiIBorderTouched, &mpiStat);
                }
                
                // MPI: Check if another Process wants to transmit a Border-Movement
                MPI_Iprobe(mpiIBottom, 0, MPI_COMM_WORLD, &mpiIBorderTouched, &mpiStat);
                
                // If there is Data to recieve => Recieve it
                while (mpiIBorderTouched == 1) {
                    
                    if (iDebugging == 2 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                        printf("Received from Bottom\n");
                    
                    // Recieve Stuff
                    MPI_Irecv(mpiAData, 5, MPI_INT, mpiIBottom, 0, MPI_COMM_WORLD, &mpiRRequest);
                    specialMovement(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, mpiAData, 3, iIntervalCount);
                    
                    MPI_Iprobe(mpiIBottom, 0, MPI_COMM_WORLD, &mpiIBorderTouched, &mpiStat);
                }
                
                // MPI: Check if another Process wants to transmit a Border-Movement
                MPI_Iprobe(mpiILeft, 0, MPI_COMM_WORLD, &mpiIBorderTouched, &mpiStat);
                
                // If there is Data to recieve => Recieve it
                while (mpiIBorderTouched == 1) {
                    
                    if (iDebugging == 2 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                        printf("Received from Left\n");
                    
                    // Recieve Stuff
                    MPI_Irecv(mpiAData, 5, MPI_INT, mpiILeft, 0, MPI_COMM_WORLD, &mpiRRequest);
                    specialMovement(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, mpiAData, 4, iIntervalCount);
                    
                    MPI_Iprobe(mpiILeft, 0, MPI_COMM_WORLD, &mpiIBorderTouched, &mpiStat);
                }
                
                if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                    printf("Step\n");
                
                
                // Kill whatever needs to be killed
                // Kill everything BUT Plankton!
                if ( aField[i][j][1]==iIntervalCount && aField[i][j][0] != iMarkPlankton )
                {
                    if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                        printf("Start Reset %d/%d\n",i,j);
                    
                    resetSector(aField, i, j);
                    
                    if (aField[i][j][0] == iMarkFish)
                        iAmountFish--;
                    else if (aField[i][j][0] == iMarkShark)
                        iAmountShark--;
                    
                    if (checkIfBorder(i, j)>0)
                    {
                    
                        // Delete from other processes if inside Border-Area
                        int temp = checkIfBorder(i, j);
                    
                        mpiAData[0] = i;
                        mpiAData[1] = j;
                        mpiAData[2] = -2;
                        mpiAData[3] = -2;
                        mpiAData[4] = iMarkEmpty;
                    
                        // Sent Data to the Process above
                        if (temp == 1)
                        {
                            MPI_Isend(mpiAData, 5, MPI_INT , mpiITop, 0, MPI_COMM_WORLD, &mpiRRequest);
                            MPI_Wait(&mpiRRequest,&mpiStat);
                        }
                    
                        if (temp == 2)
                        {
                            MPI_Isend(mpiAData, 5, MPI_INT , mpiIRight, 0, MPI_COMM_WORLD, &mpiRRequest);
                            MPI_Wait(&mpiRRequest,&mpiStat);
                        }
                        
                        if (temp == 3)
                        {
                            MPI_Isend(mpiAData, 5, MPI_INT , mpiIBottom, 0, MPI_COMM_WORLD, &mpiRRequest);
                            MPI_Wait(&mpiRRequest,&mpiStat);
                        }
                    
                        if (temp == 4)
                        {
                            MPI_Isend(mpiAData, 5, MPI_INT , mpiILeft, 0, MPI_COMM_WORLD, &mpiRRequest);
                            MPI_Wait(&mpiRRequest,&mpiStat);
                        }
                    }
                    
                    if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                        printf("Ended Reset %d/%d\n",i,j);
                    
                }
                
                
                // (aField[i][j][2]==0) := This means the field is touched for the first time this round
                if ( (aField[i][j][0] > 0) && (aField[i][j][2]==0) )
                {
                    // Increase Amount of processed Items; If this state is reached at least one Item will be processed!
                    iAmountProcessed++;
                    
                    if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                        printf("Found something\n");
                    
                    // Save Old Value for temporary Purpose
                    iOldValue = aField[i][j][0];
                    
                    // ---------------------------------------------
                    // ##             Shark Management            ##
                    // ---------------------------------------------
                    if (iOldValue == iMarkShark)
                    {
                        // Cache Values
                        y=i;
                        x=j;
                        
                        if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                            printf("Found a Shark (%d/%d)\n",x,y);
                        
                        if (checkIfFishAround(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, &y, &x) == 1) {
                            /*
                             # Fish around!
                             */
                            
                            // Over here we get 2 Sharks processed at the end of the if-Term so we need another Incriesing of iAmountProcessed
                            iAmountProcessed++;
                            
                            if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                                printf("Fish around\n");
                            
                            if (checkIfBorder(y, x)>0)
                            {
                                if (iDebugging == 2 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                                    printf("Border1\n");
                                
                                int temp = checkIfBorder(y, x);
                                int bOtherProcess = checkIfOtherProcess(y, x);
                                mpiAData[0] = i;
                                mpiAData[1] = j;
                                mpiAData[2] = y;
                                mpiAData[3] = x;
                                mpiAData[4] = iMarkShark;
                                
                                // Sent Data to the Process above
                                if (temp == 1)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiITop, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    // If Element moves to other Process; Element in Border needs to be created
                                    if (bOtherProcess > 0)
                                        aBorderTop[x]=iMarkShark;
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                                
                                if (temp == 2)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiIRight, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    // If Element moves to other Process; Element in Border needs to be created
                                    if (bOtherProcess > 0)
                                        aBorderRight[y]=iMarkShark;
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                                
                                if (temp == 3)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiIBottom, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    // If Element moves to other Process; Element in Border needs to be created
                                    if (bOtherProcess > 0)
                                        aBorderBottom[x]=iMarkShark;
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                                
                                if (temp == 4)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiILeft, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    // If Element moves to other Process; Element in Border needs to be created
                                    if (bOtherProcess > 0)
                                        aBorderLeft[y]=iMarkShark;
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                            }
                            
                            // Values leaving the Border-Area
                            if (checkIfBorder(i, j)>0 && checkIfBorder(i+y, j+x)==0)
                            {
                                if (iDebugging == 2 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                                    printf("Border3\n");
                                
                                int temp = checkIfBorder(i+y, j+x);
                                mpiAData[0] = i;
                                mpiAData[1] = j;
                                mpiAData[2] = -2;
                                mpiAData[3] = -2;
                                mpiAData[4] = iMarkShark;
                                
                                // Sent Data to the Process above
                                if (temp == 1)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiITop, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                                
                                if (temp == 2)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiIRight, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                                
                                if (temp == 3)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiIBottom, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                                
                                if (temp == 4)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiILeft, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                            }
                            
                            if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                                printf("Before Move 1\n");
                            
                            if (checkIfOtherProcess(y, x)==0) {
                                // Move Shark to the fish's position
                                moveValues(aField,i,j,y,x);
                            } else {
                                resetSector(aField, i, j);
                            }
                            
                            if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                                printf("After Move 1\n");
                            
                            iAmountFish--; // Decrease Amount of fishs
                            
                            iRnds = 0;
                            // Get position for NEW shark
                            do {
                                rndMove(&y, &x);
                                iRnds++;
                            } while ((iRnds < iRndToZero) && ( (x!=0) || (y!=0) ) && checkIfFree(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, i+y, j+x)==0 );
                            
                            if (iRnds < iRndToZero)
                            {
                            
                                if (checkIfBorder(i+y, j+x)>0)
                                {
                                    if (iDebugging == 2 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                                        printf("Border2\n");
                                
                                    int temp = checkIfBorder(i+y, j+x);
                                    int bOtherProcess = checkIfOtherProcess(i+y, j+x);
                                    mpiAData[0] = -2;
                                    mpiAData[1] = -2;
                                    mpiAData[2] = i+y;
                                    mpiAData[3] = j+x;
                                    mpiAData[4] = iMarkShark;
                                
                                    // Sent Data to the Process above
                                    if (temp == 1)
                                    {
                                        MPI_Isend(mpiAData, 5, MPI_INT , mpiITop, 0, MPI_COMM_WORLD, &mpiRRequest);
                                        // If Element moves to other Process; Element in Border needs to be created
                                        if (bOtherProcess > 0)
                                            aBorderTop[x]=iMarkShark;
                                        MPI_Wait(&mpiRRequest,&mpiStat);
                                    }
                                    
                                    if (temp == 2)
                                    {
                                        MPI_Isend(mpiAData, 5, MPI_INT , mpiIRight, 0, MPI_COMM_WORLD, &mpiRRequest);
                                        // If Element moves to other Process; Element in Border needs to be created
                                        if (bOtherProcess > 0)
                                            aBorderRight[y]=iMarkShark;
                                        MPI_Wait(&mpiRRequest,&mpiStat);
                                    }
                                    
                                    if (temp == 3)
                                    {
                                        MPI_Isend(mpiAData, 5, MPI_INT , mpiIBottom, 0, MPI_COMM_WORLD, &mpiRRequest);
                                        // If Element moves to other Process; Element in Border needs to be created
                                        if (bOtherProcess > 0)
                                            aBorderBottom[x]=iMarkShark;
                                        MPI_Wait(&mpiRRequest,&mpiStat);
                                    }
                                    
                                    if (temp == 4)
                                    {
                                        MPI_Isend(mpiAData, 5, MPI_INT , mpiILeft, 0, MPI_COMM_WORLD, &mpiRRequest);
                                        // If Element moves to other Process; Element in Border needs to be created
                                        if (bOtherProcess > 0)
                                            aBorderLeft[y]=iMarkShark;
                                        MPI_Wait(&mpiRRequest,&mpiStat);
                                    }
                                }
                            
                                // Only perform setShark if not in other Process!
                                if (checkIfOtherProcess(i+y, j+x)==0) {
                                    setShark(aField, i+y, j+x, iIntervalCount);
                                }
                            }
                        
                        } else {
                            /*
                             # No Fish around -> Normal Movement
                             */
                            
                            if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                                printf("No Fish around\n");
                            
                            iRnds=0;
                            // Get new position for shark
                            do {
                                rndMove(&y, &x);
                                iRnds++;
                            } while ( (iRnds < iRndToZero) && ( (x!=0) || (y!=0) ) && checkIfFree(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, i+y, j+x)==0);
                            
                            if (iRnds >= iRndToZero)
                            {
                                x=0;
                                y=0;
                            }
                            
                            if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                                printf("Found valid new Position (%d/%d)\n",i+y,j+x);
                            
                            if (checkIfBorder(i+y, j+x)>0)
                            {
                                if (iDebugging == 2 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                                    printf("Border3\n");
                                
                                int temp = checkIfBorder(i+y, j+x);
                                int bOtherProcess = checkIfOtherProcess(i+y, j+x);
                                mpiAData[0] = i;
                                mpiAData[1] = j;
                                mpiAData[2] = i+y;
                                mpiAData[3] = j+x;
                                mpiAData[4] = iMarkShark;
                                
                                // Sent Data to the Process above
                                if (temp == 1)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiITop, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    // If Element moves to other Process; Element in Border needs to be created
                                    if (bOtherProcess > 0)
                                        aBorderTop[x]=iMarkShark;
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                                
                                if (temp == 2)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiIRight, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    // If Element moves to other Process; Element in Border needs to be created
                                    if (bOtherProcess > 0)
                                        aBorderRight[y]=iMarkShark;
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                                
                                if (temp == 3)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiIBottom, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    // If Element moves to other Process; Element in Border needs to be created
                                    if (bOtherProcess > 0)
                                        aBorderBottom[x]=iMarkShark;
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                                
                                if (temp == 4)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiILeft, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    // If Element moves to other Process; Element in Border needs to be created
                                    if (bOtherProcess > 0)
                                        aBorderLeft[y]=iMarkShark;
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                            }
                            
                            // Values leaving the Border-Area
                            if (checkIfBorder(i, j)>0 && checkIfBorder(i+y, j+x)==0)
                            {
                                if (iDebugging == 2 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                                    printf("Border3\n");
                                
                                int temp = checkIfBorder(i+y, j+x);
                                mpiAData[0] = i;
                                mpiAData[1] = j;
                                mpiAData[2] = -2;
                                mpiAData[3] = -2;
                                mpiAData[4] = iMarkShark;
                                
                                // Sent Data to the Process above
                                if (temp == 1)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiITop, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                                
                                if (temp == 2)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiIRight, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                                
                                if (temp == 3)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiIBottom, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                                
                                if (temp == 4)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiILeft, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                            }
                            
                            if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                                printf("Before Move 2\n");
                            
                            if (checkIfOtherProcess(i+y, j+x)==0) {
                                // Set Shark to new position
                                moveValues(aField,i,j,i+y,j+x);
                            } else {
                                resetSector(aField, i, j);
                            }

                            if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                                printf("After Move 2\n");
                        }
                        
                    // ---------------------------------------------
                    // ##             Fish Management            ##
                    // ---------------------------------------------
                    } else if (iOldValue == iMarkFish) {
                        
                        if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                            printf("Found Fish (%d/%d)\n",j,i);
                        
                        iRnds=0;
                        do {
                            rndMove(&y, &x);
                            iRnds++;
                        } while ( (iRnds < iRndToZero) && ( (x!=0) || (y!=0) ) && checkIfFree(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, i+y, j+x)==0);
                        
                        if (iRnds >= iRndToZero)
                        {
                            x=0;
                            y=0;
                        }
                        
                        if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                            printf("Found valid new Position (%d/%d)\n",i+y,j+x);

                        if (checkIfBorder(i+y, j+x)>0)
                        {
                            if (iDebugging == 2 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                                printf("Border4\n");
                            
                            int temp = checkIfBorder(i+y, j+x);
                            int bOtherProcess = checkIfOtherProcess(i+y, j+x);
                            mpiAData[0] = i;
                            mpiAData[1] = j;
                            mpiAData[2] = i+y;
                            mpiAData[3] = j+x;
                            mpiAData[4] = iMarkFish;
                            
                            // Sent Data to the Process above
                            if (temp == 1)
                            {
                                MPI_Isend(mpiAData, 5, MPI_INT , mpiITop, 0, MPI_COMM_WORLD, &mpiRRequest);
                                // If Element moves to other Process; Element in Border needs to be created
                                if (bOtherProcess > 0)
                                    aBorderTop[x]=iMarkFish;
                                MPI_Wait(&mpiRRequest,&mpiStat);
                            }
                            
                            if (temp == 2)
                            {
                                MPI_Isend(mpiAData, 5, MPI_INT , mpiIRight, 0, MPI_COMM_WORLD, &mpiRRequest);
                                // If Element moves to other Process; Element in Border needs to be created
                                if (bOtherProcess > 0)
                                    aBorderRight[y]=iMarkFish;
                                MPI_Wait(&mpiRRequest,&mpiStat);
                            }
                            
                            if (temp == 3)
                            {
                                MPI_Isend(mpiAData, 5, MPI_INT , mpiIBottom, 0, MPI_COMM_WORLD, &mpiRRequest);
                                // If Element moves to other Process; Element in Border needs to be created
                                if (bOtherProcess > 0)
                                    aBorderBottom[x]=iMarkFish;
                                MPI_Wait(&mpiRRequest,&mpiStat);
                            }
                            
                            if (temp == 4)
                            {
                                MPI_Isend(mpiAData, 5, MPI_INT , mpiILeft, 0, MPI_COMM_WORLD, &mpiRRequest);
                                // If Element moves to other Process; Element in Border needs to be created
                                if (bOtherProcess > 0)
                                    aBorderLeft[y]=iMarkFish;
                                MPI_Wait(&mpiRRequest,&mpiStat);
                            }
                        }
                        
                        // Values leaving the Border-Area
                        if (checkIfBorder(i, j)>0 && checkIfBorder(i+y, j+x)==0)
                        {
                            if (iDebugging == 2 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                                printf("Border3\n");
                            
                            int temp = checkIfBorder(i+y, j+x);
                            mpiAData[0] = i;
                            mpiAData[1] = j;
                            mpiAData[2] = -2;
                            mpiAData[3] = -2;
                            mpiAData[4] = iMarkFish;
                            
                            // Sent Data to the Process above
                            if (temp == 1)
                            {
                                MPI_Isend(mpiAData, 5, MPI_INT , mpiITop, 0, MPI_COMM_WORLD, &mpiRRequest);
                                MPI_Wait(&mpiRRequest,&mpiStat);
                            }
                            
                            if (temp == 2)
                            {
                                MPI_Isend(mpiAData, 5, MPI_INT , mpiIRight, 0, MPI_COMM_WORLD, &mpiRRequest);
                                MPI_Wait(&mpiRRequest,&mpiStat);
                            }
                            
                            if (temp == 3)
                            {
                                MPI_Isend(mpiAData, 5, MPI_INT , mpiIBottom, 0, MPI_COMM_WORLD, &mpiRRequest);
                                MPI_Wait(&mpiRRequest,&mpiStat);
                            }
                            
                            if (temp == 4)
                            {
                                MPI_Isend(mpiAData, 5, MPI_INT , mpiILeft, 0, MPI_COMM_WORLD, &mpiRRequest);
                                MPI_Wait(&mpiRRequest,&mpiStat);
                            }
                        }
                        
                        if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                            printf("Before Move 3\n");
                        
                        if (checkIfOtherProcess(i+y, j+x)==0) {
                            moveValues(aField,i,j,i+y,j+x);
                        } else {
                            resetSector(aField, i, j);
                        }
                        
                        if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                            printf("After Move 3\n");
                    
                    // ---------------------------------------------
                    // ##             Plankton Management          ##
                    // ---------------------------------------------
                    } else if ( (iOldValue == iMarkPlankton) && (aField[i][j][1] == iIntervalCount) ) {
                        
                        if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                            printf("Found Plankton and Plankton should be respawned (%d/%d)\n",j,i);
                        
                        y=i;
                        x=j;
                        
                        iRnds=0;
                        // Get Position for new Plankton around
                        do {
                            rndMove(&y, &x);
                            iRnds++;
                        } while ( (iRnds < iRndToZero) && ( (x!=0) || (y!=0) ) && checkIfFree(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, i+y, j+x)==0);
                        
                        if (iRnds >= iRndToZero)
                        {
                            x=0;
                            y=0;
                        }
                        
                        if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                            printf("Found valid new Position (%d/%d)\n",i+y,j+x);
                        
                        // If no new Plankton will be set ignore the rest of the procedure;
                        // It's for the case if Plankton is surrounded by other Elements
                        if ( (x!=0) || (y!=0) )
                        {
                        
                            if (checkIfBorder(i+y, j+x)==1)
                            {
                                if (iDebugging == 2 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                                printf("Border5\n");
                            
                                int temp = checkIfBorder(i+y, j+x);
                                int bOtherProcess = checkIfOtherProcess(i+y, j+x);
                                mpiAData[0] = -2;
                                mpiAData[1] = -2;
                                mpiAData[2] = i+y;
                                mpiAData[3] = j+x;
                                mpiAData[4] = iMarkPlankton;
                            
                                // Sent Data to the Process above
                                if (temp == 1)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiITop, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    if (bOtherProcess > 0)
                                        aBorderTop[x]=iMarkPlankton;
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                                
                                if (temp == 2)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiIRight, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    if (bOtherProcess > 0)
                                        aBorderRight[y]=iMarkPlankton;
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                                
                                if (temp == 3)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiIBottom, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    if (bOtherProcess > 0)
                                        aBorderBottom[x]=iMarkPlankton;
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                                
                                if (temp == 4)
                                {
                                    MPI_Isend(mpiAData, 5, MPI_INT , mpiILeft, 0, MPI_COMM_WORLD, &mpiRRequest);
                                    if (bOtherProcess > 0)
                                        aBorderLeft[y]=iMarkPlankton;
                                    MPI_Wait(&mpiRRequest,&mpiStat);
                                }
                            }
                        
                            if (checkIfOtherProcess(i+y, j+x)==0) {
                                // Set new Plankton
                                setPlankton(aField, (i+y), (j+x), iIntervalCount);
                            }
                        }
                    }
                }
                
                if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1) && iRnds > 0)
                    printf("No of Rnds: %d\n", iRnds);
                
                iRnds = 0;
                
                iAmountAll = iAmountFish + iAmountShark + iAmountPlankton;
                
            }
        }
        
        // Initiate iAmountAll after First Run
        iAmountAll = iAmountFish + iAmountShark + iAmountPlankton;
        
        if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
            printf("Amounts2: %d/%d\n",iAmountAll,iAmountProcessed);
        
        if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
            printf("Before Reset\n");
        
        // Reset Touched for all Sectors
        resetTouched(aField);
        
        if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
            printf("After Reset\n");
        
        // ---------------
        // Print the actual field
        // printArray(aField,iFieldSizeY,iFieldSizeX);
        // ---------------
        writeArray(iIntervalCount,mpiIMe,aField,iFieldSizeY,iFieldSizeX,mpiISize);
        
        if (iDebugging > 0 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
            printf("%d : Got to Barrier for %d time\n", mpiIMe, iIntervalCount);
        
        // ---------------
        // Synconisation -> All Processes are running syncron
        // ---------------
        MPI_Barrier(MPI_COMM_WORLD);
        
        
        // Receive the Rest Data and add it to the Output
        // MPI: Check if another Process wants to transmit a Border-Movement
        MPI_Iprobe(mpiITop, 0, MPI_COMM_WORLD, &mpiIBorderTouched, &mpiStat);
        
        // If there is Data to recieve => Recieve it
        while (mpiIBorderTouched == 1) {
            
            if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                printf("Recieved Something from Top\n");
            
            // Recieve Stuff
            MPI_Irecv(mpiAData, 5, MPI_INT, mpiITop, 0, MPI_COMM_WORLD, &mpiRRequest);
            specialMovement(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, mpiAData, 1, iIntervalCount);
            
            // Write Additional Output to File
            // Check if Border-Area
            if (mpiAData[2] == iFieldSizeY)
            {
                addToOutput(iIntervalCount,mpiIMe,mpiAData[3],0,mpiAData[4]);
                printf("%d/%d : I Added something", mpiIMe, iIntervalCount);
            }
            MPI_Iprobe(mpiITop, 0, MPI_COMM_WORLD, &mpiIBorderTouched, &mpiStat);
        }
        
        // MPI: Check if another Process wants to transmit a Border-Movement
        MPI_Iprobe(mpiIRight, 0, MPI_COMM_WORLD, &mpiIBorderTouched, &mpiStat);
        
        // If there is Data to recieve => Recieve it
        while (mpiIBorderTouched == 1) {
            
            if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                printf("Recieved Something from Right\n");
            
            // Recieve Stuff
            MPI_Irecv(mpiAData, 5, MPI_INT, mpiIRight, 0, MPI_COMM_WORLD, &mpiRRequest);
            specialMovement(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, mpiAData, 2, iIntervalCount);
            // Write Additional Output to File
            // Check if Border-Area
            if (mpiAData[3] == iFieldSizeX)
            {
                addToOutput(iIntervalCount,mpiIMe,0,mpiAData[2],mpiAData[4]);
                printf("%d/%d : I Added something", mpiIMe, iIntervalCount);
            }
            MPI_Iprobe(mpiIRight, 0, MPI_COMM_WORLD, &mpiIBorderTouched, &mpiStat);
        }
        
        // MPI: Check if another Process wants to transmit a Border-Movement
        MPI_Iprobe(mpiIBottom, 0, MPI_COMM_WORLD, &mpiIBorderTouched, &mpiStat);
        
        // If there is Data to recieve => Recieve it
        while (mpiIBorderTouched == 1) {
            
            if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                printf("Recieved Something from Bottom\n");
            
            // Recieve Stuff
            MPI_Irecv(mpiAData, 5, MPI_INT, mpiIBottom, 0, MPI_COMM_WORLD, &mpiRRequest);
            specialMovement(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, mpiAData, 3, iIntervalCount);
            // Write Additional Output to File
            // Check if Border-Area
            if (mpiAData[2] == -1)
            {
                addToOutput(iIntervalCount,mpiIMe,mpiAData[3],iFieldSizeY,mpiAData[4]);
                printf("%d/%d : I Added something", mpiIMe, iIntervalCount);
            }
            MPI_Iprobe(mpiIBottom, 0, MPI_COMM_WORLD, &mpiIBorderTouched, &mpiStat);
        }
        
        // MPI: Check if another Process wants to transmit a Border-Movement
        MPI_Iprobe(mpiILeft, 0, MPI_COMM_WORLD, &mpiIBorderTouched, &mpiStat);
        
        // If there is Data to recieve => Recieve it
        while (mpiIBorderTouched == 1) {
            
            if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
                printf("Recieved Something from Right\n");
            
            // Recieve Stuff
            MPI_Irecv(mpiAData, 5, MPI_INT, mpiILeft, 0, MPI_COMM_WORLD, &mpiRRequest);
            specialMovement(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft, mpiAData, 4, iIntervalCount);
            // Write Additional Output to File
            // Check if Border-Area
            if (mpiAData[3] == -1)
            {
                addToOutput(iIntervalCount,mpiIMe,iFieldSizeX,mpiAData[2],mpiAData[4]);
                printf("%d/%d : I Added something", mpiIMe, iIntervalCount);
            }
            MPI_Iprobe(mpiILeft, 0, MPI_COMM_WORLD, &mpiIBorderTouched, &mpiStat);
        }
        
    }
    
    if (iDebugging == 1 && (mpiIMe == iDebugProcess || iDebugProcess == -1))
        printf("Start Finishing\n");
    
    //Clear Memory
    freeMemory(aField, aBorderTop, aBorderRight, aBorderBottom, aBorderLeft);
    
    // Finalize MPI-Session
    MPI_Finalize();
    
    cPrgend=clock();
    
    writeTimeToOutput(mpiIMe, cPrgstart, cPrgend);
}

